#! /bin/bash

# Lizenz / license:
# GPL v3 or later: http://www.gnu.org/licenses/gpl-3.0

# CHANGELOG
# Hauke Laging, hauke.laging (emailzeichen) openpgp-schulungen.de
#       XMPP: gnupg-support@jabber.org (OTR: http://www.hauke-laging.de/sicherheit/#mein_otr)
# 2.1, 05.01.2015, Hauke Laging
VERSION=2.1

scripts=(collect_entropy.sh key-generation.sh policy-url.sh user-ids.sh set_time_back.sh import_unix.sh install-distro-packages.sh start_keygen.sh)
DIGEST_TYPE=SHA1
SCRIPTS_DIR_NAME="scripts"

combined_hash_only=no

if [ --combined_hash_only = "$1" ]; then
	combined_hash_only=yes
fi

echo "Vorher den openpgp-Ordner aktualisieren!
" >&2

i=0
file="start.sh"
if ! [ -f "$file" ]; then
	echo "Die Datei '${file}' existiert nicht; Abbruch."
	exit 3
fi
hash_value="$(gpg --print-md "$DIGEST_TYPE" <"$file")"
if [ 50 -ne "${#hash_value}" ]; then
	echo "Fehler bei der Hashbestimmung der Datei '${file}'; Abbruch."
	exit 3
fi
script_hash[$i]="$hash_value"
((i++))
if [ no = "$combined_hash_only" ]; then
	echo "${hash_value}   ${DIGEST_TYPE}   ${file}"
fi
for filename in "${scripts[@]}"; do
	file="${SCRIPTS_DIR_NAME}/${filename}"
	if ! [ -f "$file" ]; then
		echo "Die Datei '${file}' existiert nicht; Abbruch."
		exit 3
	fi
	hash_value="$(gpg --print-md "$DIGEST_TYPE" <"$file")"
	if [ 50 -ne "${#hash_value}" ]; then
		echo "Fehler bei der Hashbestimmung der Datei '${file}'; Abbruch."
		exit 3
	fi
	script_hash[$i]="$hash_value"
	((i++))
	if [ no = "$combined_hash_only" ]; then
		echo "${hash_value}   ${DIGEST_TYPE}   ${file}"
	fi
done

if [ no = "$combined_hash_only" ]; then
	echo
	echo "Gesamthash (Hashes in der obigen Formatierung, getrennt durch ein Leerzeichen):"
fi
# set -x
# echo -n "${script_hash[*]}" | wc
echo -n "${script_hash[*]}" | gpg --print-md "$DIGEST_TYPE"

# COMMENT: Just in case the copying has destroyed the executable rights
chmod a+x "${scripts[@]}" &>/dev/null
