#!/bin/bash

# Lizenz / license:
# GPL v3 or later: http://www.gnu.org/licenses/gpl-3.0

# Documentation:
# http://www.openpgp-schulungen.de/scripte/ (German)

# This script determines the distribution it is running under and searches for a directory 
# ($PACKAGE_ROOT_DIR_NAME) which contains additional packages for one or more distributions. 
# If there are packages for the respective distribution then they get installed.

# CHANGELOG
# Hauke Laging, hauke.laging (emailzeichen) openpgp-schulungen.de
#	XMPP: gnupg-support@jabber.org (OTR: http://www.hauke-laging.de/sicherheit/#mein_otr)
# 1.1, 27.10.2014, Hauke Laging
VERSION=1.1

# TODO
# distro match per Parameter erzwingen

DISTRO_NAME[0]="Knoppix 7.2 (CD (Microknoppix), 32 Bit)"

DISTRO_UNAME_STRING[0]="Linux Microknoppix 3.9.6 #25 SMP PREEMPT Sat Jun 15 15:27:01 CEST 2013 i686 GNU/Linux"

DISTRO_SUBPATH[0]="knoppix-7.2/i386/"

DISTRO_PACKAGE_MANAGER[0]="dpkg"

E_ROOTDIRNOTFOUND="3"
E_DISTRONOTFOUND="4"
E_DISTRODIRNOTFOUND="5"

PACKAGE_ROOT_DIR_NAME="openpgp-distro-packages"
FAKE_MATCH="no"
FAKE_INDEX="0"

package_root_dir_path= # set in find_package_root_dir ()

startpath="${0%/*}"
test -n "$startpath" -a -d "$startpath" || exit 1
startpath="$(readlink --no-newline -f "$startpath"; echo x)"
startpath="${startpath%x}"
distro_match="no"
yes_char="y"
no_char="n"

uname_string=
distro_count=
distro_index=

cd "$startpath" || exit 1

yesno () {
	test $# -ne 1 && { echo "internal error in yesno(); aborting."; exit 1; }
	local text="$1" answer
	while true; do
		echo -n "${text} (${yes_char}/${no_char}): "
		read answer
		if [ "$yes_char" = "$answer" -o "$no_char" = "$answer" ]; then
			if [ "$yes_char" = "$answer" ]; then
				return 0
			else
				return 1
			fi
		else
			echo "Falsche Eingabe; bitte noch mal..."
			echo
			echo
			echo
		fi
		
	done
} # yesno ()
find_package_root_dir () {
	while true; do
		if [ -d "$PACKAGE_ROOT_DIR_NAME" ]; then
			package_root_dir_path="${PWD}/${PACKAGE_ROOT_DIR_NAME}"
			return 0
		else
			if [ "/" = "$PWD" ]; then
				return 1
			else
				cd ..
			fi
		fi
	done
} # find_package_root_dir ()

select_manually () {
	local answer i cont
	while true; do
		clear
		echo "Enter the index of the distro packages you want to install."
		echo
		echo "If you want to skip this (install nothing) just press return."
		echo
		for((i=0;i<distro_count;i++)); do
			echo "${i} : ${DISTRO_NAME[i]}"
		done
		read answer
		if [ -z "$answer" ]; then
			exit "$E_DISTRONOTFOUND"
		fi
		if [[ "$answer" =~ ^([0-9]|[1-9][0-9]*)$ && "$answer" -lt "$distro_count" ]]; then
			break
		else
			echo "Your input was incorrect, try again."
			echo "Press return to continue..."
			read cont
		fi
	done
} # select_manually ()

if ! find_package_root_dir; then
	echo "No directory '${PACKAGE_ROOT_DIR_NAME}' found; aborting."
	exit "$E_ROOTDIRNOTFOUND"
fi
echo "Found directory at path '${package_root_dir_path}'."

uname_string="$(uname -a)"
distro_count="${#DISTRO_UNAME_STRING[@]}"

for((i=0;i<distro_count;i++)); do
	if [ "x${DISTRO_UNAME_STRING[i]}" = "x$uname_string" ]; then
		distro_index=$i
		echo "Distribution identified:"
		echo "     ${DISTRO_NAME[distro_index]}"
		echo
		if ! [ -d "${package_root_dir_path}/${DISTRO_SUBPATH[distro_index]}" ]; then
			echo "Package directory for this distribution not found; aborting."
			exit "$E_DISTRODIRNOTFOUND"
		fi
		distro_match="yes"
		break
	fi
done
if [ "no" = "$FAKE_MATCH" -a "yes" != "$distro_match" ]; then
	echo "The running distribution was not recognized."
	echo
	if yesno "Do you want to select a package group manually?"; then
		select_manually
	else
		exit "$E_DISTRONOTFOUND"
	fi
fi
if [ "yes" = "$FAKE_MATCH" ]; then
	distro_index="$FAKE_INDEX"
	DISTRO_PACKAGE_MANAGER[distro_index]="dpkg"
fi

if [ "no" = "$FAKE_MATCH" ]; then
	cd "${package_root_dir_path}/${DISTRO_SUBPATH[distro_index]}" || exit 1
fi

case "${DISTRO_PACKAGE_MANAGER[distro_index]}" in
	dpkg)
		if [ "no" = "$FAKE_MATCH" ]; then
			dpkg -i *
		else
			echo "FAKE_MATCH: dpkg -i *"
		fi
	;;
	rpm)
		if [ "no" = "$FAKE_MATCH" ]; then
			rpm -i *
		else
			echo "FAKE_MATCH: rpm -i *"
		fi
	;;
	*)
		echo "Internal error (package manager selection); aborting."
		exit 1
	;;
esac

