#! /bin/bash

# This script edits (or deletes) the policy URL in the file $SOURCE_FILE in
# the current working directory.

# Lizenz / license:
# GPL v3 or later: http://www.gnu.org/licenses/gpl-3.0

# CHANGELOG
# 1.8, 05.01.2015, Hauke Laging, www.openpgp-schulungen.de
VERSION=1.8

SOURCE_FILE=gpg.conf
TMP_FILE=gpg.conf.tmp
DEFAULT_DIR_URL="http://www.openpgp-schulungen.de/users/"
DEFAULT_FILE="__SHORTID____policy.1.html"
URL_VAR_STRING=__SHORTID__
DEFAULT_SHOW_URL="http://www.meine.domain/openpgp/"
EXIT_CODE_NOPOLICYURL="5"
CONFIGFILE_DEFAULT_URL="http://www.openpgp-schulungen.de/users/__SHORTID____policy.1.html"
MODIFY_STRING="MODIFY"
MODIFY_LEGACY_STRING="ANPASSEN"
REMOVE_STRING="REMOVE"

declare -A STRINGS
declare -A STRINGS_en
declare -A STRINGS_de
STRINGS_en["USAGE"]="Usage"
STRINGS_de["USAGE"]="Aufruf"

shopt -s extglob

select_language () {
local version=1.0
	local varname key tmp
	use_language=
	for varname in LANGUAGE LC_ALL LC_MESSAGES LANG; do
		if [ -n "${!varname}" ]; then
			use_language="${!varname}"
			use_language="${use_language%%_*}"
			break
		fi
	done
	if [ -z "$use_language" ]; then
		use_language="en"
	fi
	for key in "${!STRINGS_en[@]}"; do
		eval tmp="\"\${STRINGS_${use_language}[$key]}\""
		if [ -n "$tmp" ]; then
			STRINGS[$key]="$tmp"
		else
			STRINGS[$key]="${STRINGS_en[key]}"
		fi
	done
} # select_language ()
select_language

default_url_user="$DEFAULT_SHOW_URL" # the URL shown to the user
yes_char="j"
no_char="n"
webspace_available=
file_value="no" # is the value in the config file shown? (or the script default or user input)
file_value_hint=

usage () {
	echo "${STRINGS["USAGE"]}:"
	echo "'$0' [--no-policy-url]"
} # usage ()

if [ "--help" = "$1" ]; then
	usage
	exit
fi

yesno () {
	test $# -ne 1 && { echo "internal error in yesno(); aborting."; exit 1; }
	local text="$1" answer
	while true; do
		echo -n "${text} (${yes_char}/${no_char}): "
		read answer
		if [ "$yes_char" = "$answer" -o "$no_char" = "$answer" ]; then
			if [ "$yes_char" = "$answer" ]; then
				return 0
			else
				return 1
			fi
		else
			echo "Falsche Eingabe; bitte noch mal..."
			echo
			echo
			echo
		fi
		
	done
} # yesno ()

clear () {   
	if [ yes = "$DEBUG" ]; then
	# im Debug-Modus clear ignorieren; nervt nur
		echo "######################################"
		echo "DEBUG-Modus: clear ignoriert"
		echo "######################################"
	else
		command clear
	fi
} # clear ()

no_policy_url () {
	awk '/^#.*set-policy-url/ {sub("_REMOVE_","_REM_OVE_");sub("'"${MODIFY_STRING}"'","");}; {print;}' "$SOURCE_FILE" > "$TMP_FILE" && mv -f "$TMP_FILE" "$SOURCE_FILE"
} # no_policy_url ()

convert_legacy_file () {
	sed -i s/"${MODIFY_LEGACY_STRING}"/"${MODIFY_STRING}"/ "$SOURCE_FILE"
} # convert_legacy_file ()

convert_legacy_file

if [ "--no-policy-url" = "$1" ]; then
	no_policy_url
	exit 0
fi

source_path="${PWD}/${SOURCE_FILE}"
length=${#source_path}
if [ 72 -gt "$length" ]; then
	padding_length=$(((72-length)/2))
	padding="$(printf %${padding_length}s "")"
else
	padding=
fi

if ! [ -f "$source_path" -a -r "$source_path" ]; then
	clear
	cat <<EOT
FEHLER (Abbruch)
======================================================================

Die von diesem Script zu bearbeitende Datei
${padding}${source_path}
existiert nicht oder ist nicht lesbar.

EOT
	exit 3
fi

if ! [ -f "${source_path}.ori" ]; then
	cp "$source_path" "${source_path}.ori"
fi

convert_legacy_file

if [ -f "$TMP_FILE" ]; then
	clear
	cat <<EOT
FEHLER (Abbruch)
======================================================================

Die von diesem Script zu überschreibende TMP-Datei

${PWD}${TMP_FILE}

existiert bereits. Bitte löschen Sie diese Datei und starten dieses Script 
dann neu.

EOT
	exit 3
fi

if ! [ -w . ]; then
	clear
	cat <<EOT
FEHLER (Abbruch)
======================================================================

Das aktuelle Verzeichnis ('${PWD}') ist nicht beschreibbar.

Bitte beheben Sie das Problem und starten dieses Script dann neu.

EOT
	exit 3
fi

clear
cat <<EOT
INFORMATION (kann man mit Return überspringen, muss man nicht lesen)
======================================================================

Damit andere Ihren Schlüssel sinnvoll nutzen können, müssen sie wissen, 
was von Ihrem Schlüssel zu halten ist (wie sicher er ist, wofür er gedacht 
ist usw.).

Es ist sinnvoll, die Information (URL), wo diese Information (ggf. später 
mal) zu finden ist, von Anfang an in den Schlüssel zu schreiben. Man nennt 
dieses Feature eine "policy URL".

Weiter mit Return...
EOT
read cont

clear
	cat <<EOT
AKTION ERFORDERLICH: Webspace verfügbar?
======================================================================

Verfügen Sie über eigenen Webspace, den Sie langfristig nutzen können 
(z.B. eigene Domain; Uni-Webserver)? Es ist kein großes Problem, wenn 
Ihnen die Webadresse nicht "gehört". Wenn Ihnen jemand ein Verzeichnis 
auf seiner Webseite dauerhaft zur Verfügung stellt, ist das OK.

EOT
if yesno "Eigene URL eingeben?"; then
	webspace_available="yes"
else
	echo
	if yesno "Statt dessen den Standardwert eintragen (das wird empfohlen)?"; then
		webspace_available="no"
	else
		no_policy_url
		exit "$EXIT_CODE_NOPOLICYURL"
	fi
fi

linenr_url="$(awk '/_REMOVE_set-policy-url/ {print NR " " $NF}' "$source_path")"
current_url="${linenr_url#* }"
if [ "$CONFIGFILE_DEFAULT_URL" = "$current_url" ]; then
	current_url="$default_url_user"
	file_value="no"
else
	file_value="yes"
fi
linenr="${linenr_url% *}"

if [ "no" = "$webspace_available" ]; then
	newurl="${DEFAULT_DIR_URL}${DEFAULT_FILE}"
else
	if [ "yes" != "$LESS_INFO_AND_QUESTIONS" ]; then
		clear
		cat <<EOT
INFORMATION (kann man mit Return überspringen, muss man nicht lesen)
======================================================================

Dieses Script bearbeitet die Datei
${padding}${source_path}

Es trägt die policy URL in diese Datei ein. Sie können eine URL auf ein 
Verzeichnis eingeben (der Dateiname wird automatisch ergänzt) oder das 
entsprechende Feld leer lassen (bzw. den vorgegeben Inhalt komplett 
löschen); bei einer leeren Eingabe wird der Standardwert

     ${DEFAULT_DIR_URL}

eingetragen.

Es dürfte sinnvoll sein, dass Sie dafür ein eigenes Verzeichnis verwenden, 
in der Art:
               http://www.meine.domain/openpgp/


Sie können dieses Script jederzeit durch ^C abbrechen.

Weiter mit Return...
EOT
		read cont
	fi

	if [ -n "$current_url" ]; then
		current_url_wohttp="${current_url#http?(s)://}"
		if [ x"$current_url_wohttp" != x"$current_url" ]; then
			protocol="${current_url%"$current_url_wohttp"}"
		else
			protocol=http://
		fi
		current_url_user="${protocol}${current_url_wohttp%/*}/"
	fi

	while true; do
		if [ "$file_value" = "no" ]; then
			file_value_hint=
		else
			file_value_hint=$'\nAngezeigt wird der aktuelle Wert aus der Datei:'
		fi
		clear
		cat <<EOT
AKTION ERFORDERLICH: URL eingeben
======================================================================

Bitte geben Sie die URL ein (d.h. überschreiben Sie die Beispiel-URL).

Die URL muss mit http:// oder https:// beginnen und auf / enden, 
sinnvollerweise auf /openpgp/

Löschen Sie den Inhalt (^U), um den Standardwert eintragen zu lassen
(${DEFAULT_DIR_URL})
${file_value_hint}
EOT
		read -e -i "$current_url_user" newurl_user
		file_value="no"

		# COMMENT: Auf ein paar illegale Zeichen testen; besser wäre natürlich ein 
		#	kompletter Check auf eine korrekte URL

		if [ -z "$newurl_user" ]; then
			break
		else
			if echo "${newurl_user}" | grep -qi -- '[ äöüß]'; then
				clear
				cat <<-EOT
	FEHLER
	======================================================================

	Die eingegebene URL

	'${newurl_user}'

	enthält mindestens ein illegales Zeichen.
	

	Weiter mit Return...
	EOT
				read cont
				current_url_user="$default_url_user"
				continue
			fi
			if [ "${#newurl_user}" -lt 12 ]; then
				echo "Die eingegebene URL ist zu kurz; Abbruch."
				echo "Weiter mit Return..."
				read cont
				current_url_user="$default_url_user"
				continue
			fi
			if [ x"$newurl_user" = x"${newurl_user%/}" ]; then
				echo "Die eingegebene URL endet nicht auf /"
				echo "Weiter mit Return..."
				current_url_user="${newurl_user}/"
				read cont
				continue
			fi
			if [ x"$newurl_user" = x"${DEFAULT_SHOW_URL}" ]; then
				echo "Die Beispiel-URL ist eine unzulässige Eingabe."
				echo "Weiter mit Return..."
				read cont
				current_url_user="$default_url_user"
				continue
			fi
		fi
		break
	done

	if [ -z "$newurl_user" ]; then
		newurl="${DEFAULT_DIR_URL}${DEFAULT_FILE}"
	else
		newurl="${newurl_user}${DEFAULT_FILE}"
	fi
fi

if awk -v linenr="$linenr" -v url="$newurl" \
	'NR == linenr {print "# _REMOVE_set-policy-url " url; next;} {print}' "$SOURCE_FILE" > "$TMP_FILE" && mv -f "$TMP_FILE" "$SOURCE_FILE"; then
	cat <<EOT


PROGRAMMERGEBNIS (sollte man beachten)
======================================================================

Die Konfigurationsdatei wurde angepasst.





                                ENDE
======================================================================

             Das war's. Danke und bis zum nächsten Mal!

                     www.openpgp-schulungen.de
EOT
else
	clear
	cat <<EOT
FEHLER (Abbruch)
======================================================================

Beim Versuch, die Konfigurationsdatei anzupassen, trat ein Fehler auf.

EOT
	exit 1
fi
