#! /bin/bash

# Lizenz / license:
# GPL v3 or later: http://www.gnu.org/licenses/gpl-3.0

# Documentation:
# http://www.openpgp-schulungen.de/scripte/user-ids/ (German)

# CHANGELOG
# Hauke Laging, hauke.laging (emailzeichen) openpgp-schulungen.de
#       XMPP: gnupg-support@jabber.org (OTR: http://www.hauke-laging.de/sicherheit/#mein_otr)
### $VERSION nicht vergessen!
# 1.6, 05.01.2015, Hauke Laging
VERSION=1.6

CONFIG_FILE="key-generation.conf"
SCRIPTNAME="user-ids"
DEFAULT_POLICY_URL="no"
COMMENT_POLICY_URL="Alltagsschlüssel mit sicherem Offline-Hauptschlüssel und policy URL"
COMMENT_NO_POLICY_URL="Alltagsschlüssel mit sicherem Offline-Hauptschlüssel"

declare -A STRINGS
declare -A STRINGS_en
declare -A STRINGS_de
STRINGS_en["USAGE"]="Usage"
STRINGS_de["USAGE"]="Aufruf"

select_language () {
local version=1.0
	local varname key tmp
	use_language=
	for varname in LANGUAGE LC_ALL LC_MESSAGES LANG; do
		if [ -n "${!varname}" ]; then
			use_language="${!varname}"
			use_language="${use_language%%_*}"
			break
		fi
	done
	if [ -z "$use_language" ]; then
		use_language="en"
	fi
	for key in "${!STRINGS_en[@]}"; do
		eval tmp="\"\${STRINGS_${use_language}[$key]}\""
		if [ -n "$tmp" ]; then
			STRINGS[$key]="$tmp"
		else
			STRINGS[$key]="${STRINGS_en[key]}"
		fi
	done
} # select_language ()
select_language

yes_char="j"
no_char="n"
mode=normal # normal | hisec
emails=()
comments=()
intro="yes"
policy_url="$DEFAULT_POLICY_URL"
policy_url_selection="no"

usage () {
	echo "${STRINGS["USAGE"]}:"
	echo "'$0' [--policy-url | --no-policy-url]"
} # usage ()

if [ "--help" = "$1" ]; then
	usage
	exit
fi

yesno () {
	test $# -ne 1 && { echo "internal error in yesno(); aborting."; exit 1; }
	local text="$1" answer
	while true; do
		echo -n "${text} (${yes_char}/${no_char}): "
		read answer
		if [ "$yes_char" = "$answer" -o "$no_char" = "$answer" ]; then
			if [ "$yes_char" = "$answer" ]; then
				return 0
			else
				return 1
			fi
		else
			echo "Falsche Eingabe; bitte noch mal..."
			echo
			echo
			echo
		fi
		
	done
} # yesno ()

show_uids () {
	local i count=${#emails[@]}
	for((i=1;i<=count;i++)); do
		if [ -n "${comments[i]}" ]; then
			echo "${i}.: <${emails[i]}> (${comments[i]})"
		else
			echo "${i}.: <${emails[i]}>"
		fi
	done
	if [ "$count" -gt 0 ]; then
		return 0
	else
		return 1
	fi
} # show_uids ()

backup_config () {
	if [ -f "$CONFIG_FILE" ]; then
		for((i=0;i<100;i++)); do
			if ! [ -f "$CONFIG_FILE"."$i" ]; then
				mv "$CONFIG_FILE" "$CONFIG_FILE"."$i" &&
					echo "Die existierende Konfigurationsdatei '$CONFIG_FILE' wurde in ${CONFIG_FILE}.${i} umbenannt."
					echo
				break
			fi
		done
	fi
} # backup_config ()

select_policy_url_mode () {
	cat <<EOT
AKTION ERFORDERLICH: policy URL
======================================================================

Es ist sinnvoll, eine Schlüsselrichtlinie zu verfassen, siehe

http://www.openpgp-schulungen.de/kurzinfo/schluesselqualitaet/#keypolicy

Wenn Sie das (irgendwann) tun werden, können Sie jetzt schon die URL, unter 
der diese Schlüsselrichtlinie abrufbar sein wird, in das Zertifikat (den 
öffentlichen Schlüssel) schreiben lassen, damit jeder Nutzer Ihres Schlüssels 
sehen kann, wo er diese wichtige Information findet. Das nennt man eine 
policy URL.

Da eine policy URL leider eher unauffällig ist, ist es sinnvoll, im 
Kommentarbereich des Schlüssels darauf hinzuweisen, dass es eine gibt.

EOT
	if yesno "Möchten Sie auf die policy URL hinweisen?"; then
		policy_url="yes"
	else
		policy_url="no"
	fi
	policy_url_selection="yes"
} # select_policy_url_mode ()

if ! [ -w . ]; then
	echo "Das aktuelle Verzeichnis kann nicht beschrieben werden; Abbruch."
	exit 3
fi

while [ $# -gt 0 ]; do
	case "$1" in
		--high-security)
			mode="hisec"
		;;
		--no-intro)
			intro="no"
		;;
		--policy-url)
			policy_url="yes"
			policy_url_selection="yes"
		;;
		--no-policy-url)
			policy_url="no"
			policy_url_selection="yes"
		;;
		*)
			echo "Illegaler Parameter ('${1}'); Abbruch."
			exit 2
		;;
	esac
	shift
done


clear
if [ "yes" = "$intro" -o "no" = "$LESS_INFO_AND_QUESTIONS" ]; then
	cat <<EOT
INFORMATION (kann man mit Return überspringen, muss man nicht lesen)
======================================================================

Dieses Script fragt die Daten für die User-IDs ab und schreibt sie im 
aktuellen Verzeichnis in die Datei '${CONFIG_FILE}'.

Dieses Script deckt nur einfache Fälle der Konfiguration ab. Bei 
Sonderwünschen muss die Datei direkt editiert werden.


Der Name ist bei allen User-IDs einheitlich und wird deshalb nur einmal 
abgefragt (Pseudonym ist möglich).


Weiter mit Return...

EOT
	read cont
	clear
fi

cat <<EOT
AKTION ERFORDERLICH: Ihr Name
======================================================================

Die Angaben im OpenPGP-Zertifikat sollten genauso aussehen wie die in den 
E-Mails, mit denen das Zertifikat verwendet werden soll (möglicherweise 
ist es sinnvoll, die E-Mail-Konfiguration anzupassen). Ihr E-Mail-Absender 
sieht für die Empfänger etwa so aus:

    Max Müller <mmueller@example.net> 


Geben Sie jetzt zunächst Ihren Namen ein (wie im E-Mail-Absender), die 
E-Mail-Adressen werden später abgefragt.

Beispiel: Max Müller

EOT

read -e -p "Ihr Name: " name
echo

if [ -z "$(echo $name)" ]; then
	echo "Fehlerhafter Name; Abbruch."
	exit 3
fi

if [ normal = "$mode" ]; then
	index=1

	while true; do
		clear
		echo "$name"
		if show_uids; then
			echo
			echo
		fi
		if [ "$index" -gt 1 ]; then
			echo "Wenn Sie keine weitere User-ID erzeugen möchten, "\
	"beenden Sie ohne Eingabe mit Return."
			echo
		fi
		# echo -n "${index}. E-Mail-Adresse: "
		read -e -p "${index}. E-Mail-Adresse: "	email
		if [ -z "$email" ]; then
			break
		fi
		if ! [[ $email =~ ^[a-zA-Z0-9._-]+@[a-zA-Z0-9._-]+\.[a-zA-Z]+$ ]]; then
			echo "Fehlerhafte E-Mail-Adresse; weiter mit Return..."
			read cont
			continue
		fi
		emails[index]="$email"
		echo
		echo "Kommentar zu dieser E-Mail-Adresse (meist unnötig) oder weiter mit Return"
		echo "Ein Kommentar darf keine () Klammern oder <>-Zeichen enthalten!"
		echo
		# echo -n "Kommentar: "
		read -e -p "Kommentar: " comment
		if [ -n "$comment" ]; then
			if [[ $comment =~ [\(\)\<\>] ]]; then
				echo "Ungültiges Zeichen im Kommentar; weiter mit Return..."
			fi
		fi
		comments[index]="$comment"
		((index++))
	done
	
	backup_config
	if [ "yes" != "$policy_url_selection" ]; then
		select_policy_url_mode
	fi
	
	exec 3>"$CONFIG_FILE"
	
	if [ "yes" = "$policy_url" ]; then
		comment="$COMMENT_POLICY_URL"
	else
		comment="$COMMENT_NO_POLICY_URL"
	fi

	cat >&3 <<EOT
# Konfigurationsdatei für key-generation.sh, siehe http://www.openpgp-schulungen.de/scripte/keygeneration/

uid_0_name="${name}"
uid_0_email= # Empfehlung: eine User-ID ohne E-Mail-Adresse!
uid_0_comment="${comment}"
EOT

	count=${#emails[@]}
	for((i=1;i<=count;i++)); do
		echo
		echo "uid_${i}_name=\"${name}\""
		echo "uid_${i}_email=\"${emails[i]}\""
		if [ -n "${comments[i]}" ]; then
			echo "uid_${i}_comment=\"${comments[i]}\""
		fi
	done >&3
elif [ hisec = "$mode" ]; then
	backup_config
	
	exec 3>"$CONFIG_FILE"

	cat >&3 <<EOT
# Konfigurationsdatei für key-generation.sh, siehe http://www.openpgp-schulungen.de/scripte/keygeneration/

uid_0_name="${name}"
uid_0_comment="Hochsicherheitsschlüssel, siehe policy URL; nicht für E-Mail"
EOT
else
	echo "(${SCRIPTNAME}) Interner Fehler; Abbruch"
	exit 1
fi

echo "Die Konfigurationsdatei '$CONFIG_FILE' wurde erstellt:"
echo
cat "$CONFIG_FILE"

if [ "yes" != "$LESS_INFO_AND_QUESTIONS" ]; then
	echo
	echo "Weiter mit Return"
	read cont
fi

