#! /bin/bash

# Lizenz / license:
# GPL v3 or later: http://www.gnu.org/licenses/gpl-3.0

# CHANGELOG
# Hauke Laging, hauke.laging (emailzeichen) openpgp-schulungen.de
#       XMPP: gnupg-support@jabber.org (OTR: http://www.hauke-laging.de/sicherheit/#mein_otr)
#
# environment:
# DEBUG=
# DEBUGSCRIPTS=
#
### $VERSION nicht vergessen!
# 2.5, 05.01.2015, Hauke Laging
VERSION=2.5

shopt -s nullglob

if [ "yes" = "$DEBUG" ]; then
	set -vx
fi

SCRIPT_NAME="start.sh"
SCRIPTS_DIR="scripts/"
START_KEYGEN_SCRIPT="start_keygen.sh"

startpath="${0%/*}"
test -n "$startpath" -a -d "$startpath" || exit 1
startpath="$(readlink --no-newline -f "$startpath"; echo x)"
startpath="${startpath%x}"
yes_char="j"
no_char="n"
singlestep="no"
bash_options=

usage () {
	echo "'$0' [--single-step]"
}

if [ "--help" = "$1" ]; then
	usage
	exit 0
fi

if [ "--single-step" = "$1" ]; then
	singlestep="yes"
fi

if [ "yes" = "$DEBUGSCRIPTS" ]; then
	bash_options="-vx"
fi

yesno () {
	test $# -ne 1 && { echo "internal error in yesno(); aborting."; exit 1; }
	local text="$1" answer
	while true; do
		echo -n "${text} (${yes_char}/${no_char}): "
		read answer
		if [ "$yes_char" = "$answer" -o "$no_char" = "$answer" ]; then
			if [ "$yes_char" = "$answer" ]; then
				return 0
			else
				return 1
			fi
		else
			echo "Falsche Eingabe; bitte noch mal..."
			echo
			echo
			echo
		fi
		
	done
} # yesno ()


cd "$SCRIPTS_DIR" || exit 1

if ! [ -f "${START_KEYGEN_SCRIPT}" ]; then
	echo "The key generation script does not exist; aborting."
	exit 3
fi

if [ "yes" = "$singlestep" ]; then
	bash $bash_options "./${START_KEYGEN_SCRIPT}" --single-step
else
	bash $bash_options "./${START_KEYGEN_SCRIPT}"
fi

