#! /bin/bash

# Lizenz / license:
# GPL v3 or later: http://www.gnu.org/licenses/gpl-3.0

# CHANGELOG
# Hauke Laging, hauke.laging (emailzeichen) openpgp-schulungen.de
#       XMPP: gnupg-support@jabber.org (OTR: http://www.hauke-laging.de/sicherheit/#mein_otr)
#
# environment:
# DEBUG=
# DEBUGSCRIPTS=
#
### $VERSION nicht vergessen!
# 1.1, 07.12.2015, Hauke Laging
VERSION=1.1
SCRIPT_NAME="start_keygen.sh"

shopt -s nullglob

if [ "yes" = "$DEBUG" ]; then
	set -vx
fi

USERID_SCRIPT="user-ids.sh"
POLICYURL_SCRIPT="policy-url.sh"
KEYGEN_SCRIPT="key-generation.sh"
PASSPHRASE_SCRIPT="passphrase.sh"
TIME_SCRIPT="set_time_back.sh"
RUN_FILE_TIME="keygeneration-time"
BG_ENTROPY_SCRIPT="call_bg-entropy.sh"
# USE_SCREEN=
SESSIONNAME="keygeneration"
COLLECT_ENTROPY_SCRIPT="collect_entropy.sh"
WINDOW_NAME_MAIN="main"
WINDOW_NAME_COLLECT="collect"
WINDOW_NAME_SWITCH="entropy_window_switch"
SCRIPT_SWITCH="entropy_window_switch.sh"
TIME_RUN_FILE="keygeneration-time"
LESS_INFO_AND_QUESTIONS=; export LESS_INFO_AND_QUESTIONS
PACKAGE_SCRIPT="install-distro-packages.sh"

startpath="${0%/*}"
test -n "$startpath" -a -d "$startpath" || exit 1
startpath="$(readlink --no-newline -f "$startpath"; echo x)"
startpath="${startpath%x}"
mode= # normal | hisec
start_seed="no"
yes_char="j"
no_char="n"
singlestep="no"
bash_options=
policy_url=
policy_url_parameter=

usage () {
	echo "'$0' [--single-step]"
}

if [ "--help" = "$1" ]; then
	usage
	exit 0
fi

if [ "--single-step" = "$1" ]; then
	singlestep="yes"
fi

if [ "yes" = "$DEBUGSCRIPTS" ]; then
	bash_options="-vx"
fi

yesno () {
	test $# -ne 1 && { echo "internal error in yesno(); aborting."; exit 1; }
	local text="$1" answer
	while true; do
		echo -n "${text} (${yes_char}/${no_char}): "
		read answer
		if [ "$yes_char" = "$answer" -o "$no_char" = "$answer" ]; then
			if [ "$yes_char" = "$answer" ]; then
				return 0
			else
				return 1
			fi
		else
			echo "Falsche Eingabe; bitte noch mal..."
			echo
			echo
			echo
		fi
		
	done
} # yesno ()

policy_url_help () {
	clear
	cat <<EOT | less
(Beenden der Textanzeige mit q (wie quit))
	
INFORMATION (kann man mit Return überspringen, muss man nicht lesen)
======================================================================

Damit ein hohes Maß an Sicherheit auch in der Praxis sinnvoll genutzt werden 
kann, ist Transparenz über das Sicherheitsniveau erforderlich, d.h., der 
Kommunikationspartner muss WISSEN, wie sicher etwas ist. Es hat wenig Sinn, 
mit guten Schlüsseln zu arbeiten, wenn irgendwo dann doch nur geraten wird.

Diese Schlüsselrichtlinie (policy) muss natürlich ebenfalls ein hohes 
Sicherheitsniveau haben, denn anderenfalls könnte ein Angreifer für einen 
unsicheren Schlüssel eine Richtlinie erstellen, die behauptet, dieser 
Schlüssel sei ganz besonders sicher.

Die Anforderungen in der Praxis sind also:
1) Es muss eine Schlüsselrichtlinie geben.
2) Sie muss sehr sicher sein (gegen Fälschung).
3) Der Kommunikationspartner muss wissen, dass und wo es sie gibt.
4) Der Kommunikationspartner muss wissen können, dass sie sicher ist.
5) Sie muss inhaltlich das abdecken, was den anderen interessiert.

Punkt (3) kann man mit einer policy URL erledigen: In das Zertifikat und 
ggf. auch in jede Signatur wird die URL geschrieben, unter der man die 
Schlüsselrichtlinie finden kann.

Das Dokument muss dort noch nicht abrufbar sein, um die URL einzutragen. Man 
kann eine policy URL auch nachträglich in ein Zertifikat aufnehmen, aber es 
hat sicherheitstechnische Vorteile, das von Anfang an zu machen. Außerdem 
ist es unschön, wenn man deswegen später die Kommentar-UID ändern will. Wenn 
Sie also irgendwann mal eine Schlüsselrichtlinie für diesen Schlüssel 
erstellen möchten, dann sollten Sie die URL dafür jetzt schon in das 
Zertifikat eintragen.

Es muss auch keine Website sein, die einem selber gehört; ein anderer 
Website-Betreiber kann die Signatur nicht fälschen und deshalb kaum Schaden 
anrichten. Wenn Sie auf gar keine Website Zugriff haben, können Sie Ihre 
Schlüsselrichtlinie auch kostenlos auf www.openpgp-schulungen.de ablegen. 
In dem Fall trägt dieses Script die URL automatisch ein, Sie brauchen sie 
also nicht zu wissen

Ausführliche Erklärung: http://www.openpgp-schulungen.de/glossar/keypolicy/

EOT
} # policy_url_help ()

ask_policy_url () {
	clear
	cat <<EOT
AKTION ERFORDERLICH: policy URL
======================================================================

Es ist sinnvoll, eine Schlüsselrichtlinie zu verfassen. Wenn Sie das 
(irgendwann) tun werden, können Sie jetzt schon die URL, unter der diese 
Schlüsselrichtlinie abrufbar sein wird, in das Zertifikat (den 
öffentlichen Schlüssel) schreiben lassen, damit jeder Nutzer Ihres Schlüssels 
sehen kann, wo er diese wichtige Information findet. Diese Adresse nennt man 
eine policy URL.

Da eine policy URL ((mangels Verbreitung) bei aktueller Software) leider 
eher unauffällig ist, wird dann auch im Kommentarbereich des Schlüssels 
darauf hingewiesen.

EOT
	if yesno "Soll ein längerer Text zum Thema Schlüsselrichtlinien angezeigt werden?"; then
		policy_url_help
		echo
		echo
	fi
	if yesno "Soll der Schlüssel mit Hinweis auf die Schlüsselrichtlinie erstellt werden?"; then
		policy_url="yes"
		policy_url_parameter="--policy-url"
	else
		policy_url="no"
		policy_url_parameter="--no-policy-url"
	fi
} # ask_policy_url ()

cache_progs () {
	(gpg --list-keys &>/dev/null
	cat /dev/null
	dd if=/dev/null of=/dev/null bs=512 count=1 &>/dev/null
	echo foo | awk '0' &>/dev/null
	ls / &>/dev/null
	mkdir -p --mode=700 /tmp/keygen &>/dev/null
	test -d /tmp/keygen && rmdir /tmp/keygen &>/dev/null
	touch /tmp/testfile &>/dev/null
	cp /tmp/testfile /tmp/testfile2 &>/dev/null
	rm /tmp/testfile /tmp/testfile2 &>/dev/null
	) &
} # cache_progs ()


### MAIN

cd "$startpath" || exit 1

if [ -z "$USE_SCREEN" ]; then
	if type screen &>/dev/null; then
		if screen -ls | grep -q -F ."$SESSIONNAME"; then
			echo "Es läuft bereits eine screen-Session mit dem Namen '${SESSIONNAME}'; Abbruch."
			exit 1
		fi
		USE_SCREEN=yes screen -S "$SESSIONNAME" -t "$WINDOW_NAME_MAIN" -h 1000 ./"$SCRIPT_NAME" "$@"
	else
		USE_SCREEN=no ./"$SCRIPT_NAME" "$@"
	fi
	echo "Das (Sammel-)Script ist jetzt fertig."
	exit 0
fi

cleanup () {
	if [ yes = "$USE_SCREEN" ]; then
		screen -S keygeneration -X quit
	fi
}

trap cleanup EXIT

if [ yes = "$USE_SCREEN" ]; then
	screen -t "$WINDOW_NAME_COLLECT" ./"$COLLECT_ENTROPY_SCRIPT" --no-infotext
	sleep 1
	screen -S "$SESSIONNAME" -X select "$WINDOW_NAME_MAIN"
	
	# screen -t "$WINDOW_NAME_SWITCH" ./"$SCRIPT_SWITCH"
	# screen -S "$SESSIONNAME" -X select "$WINDOW_NAME_MAIN"
fi

cache_progs
clear
cat <<EOT
AKTION ERFORDERLICH: Netzwerkverbindung prüfen
======================================================================

Falls Sie mit diesem Script hochsichere Schlüssel erstellen oder bearbeiten 
möchten, sollten Sie darauf achten, dass der Computer nicht an ein 
Netzwerkkabel angeschlossen ist, wenn Sie Ihn einschalten.

Wenn jetzt noch ein Kabel angeschlossen ist, entfernen Sie es, schalten Sie 
den Computer aus und beginnen Sie die Aktion von vorn.

Auch die Kabel von für die Schlüsselverwaltung unnötigen Geräten (Drucker 
usw.) sollten Sie entfernen.



Weiter mit Return...

EOT
read cont

if [ "yes" != "$LESS_INFO_AND_QUESTIONS" -a "no" != "$LESS_INFO_AND_QUESTIONS" ]; then
	clear
	cat <<EOT
AKTION ERFORDERLICH: Umfang der Textausgabe und Abfragen
======================================================================

Möchten Sie, dass dieses Script Meldungen ausgibt, die lediglich informativ, 
aber nicht zwingend erforderlich sind, und verzichtbare Nachfragen macht?

Dies ist typischerweise sinnvoll, wenn Sie dieses Script ohne Hilfe durch 
Dritte ausführen, aber wenn Sie in einer Schulung sitzen, sollten Sie darauf 
verzichten, um Zeit zu sparen und weil Sie die Detailnachfragen womöglich 
nicht verstehen.

EOT
	while true; do
		echo -n "Optionale Informationen ausgeben: (j)a / (n)ein: "
		read answer
		if [ j != "$answer" -a n != "$answer" ]; then
			echo "Falsche Eingabe; neuer Versuch..."
			echo
			continue
		fi
		if [ j = "$answer" ]; then
			LESS_INFO_AND_QUESTIONS="no"
		else
			LESS_INFO_AND_QUESTIONS="yes"
		fi
		echo
		break
	done
fi

# BEGIN: package install script

if [ "yes" = "$KEYGEN_NO_PACKAGES" ]; then
	:
elif [ "yes" = "$LESS_INFO_AND_QUESTIONS" ]; then
	echo "Installation der Zusatzpakete"
	echo
	echo "Jetzt müssen Sie möglicherweise das root-Passwort eingeben:"
	sudo bash $bash_options ./"$PACKAGE_SCRIPT"
else
	clear
	cat <<EOT
AKTION ERFORDERLICH: Zusatzpakete installieren
======================================================================

Möchten Sie, dass eventuell vorhandene Pakete für die gerade laufende 
Distribution (offline) nachinstalliert werden? Das Paketmanager stellt 
sicher, dass nur korrekt signierte Pakete installiert werden.

Das Fehlen benötigter Software kann den weiteren Ablauf erschweren (d.h. 
verkomplizieren und/oder verlangsamen).

EOT
	if yesno "Zusatzpakete installieren?"; then
		if [ "0" -ne "$EUID" ]; then
			echo "Jetzt müssen Sie möglicherweise das root-Passwort eingeben:"
			sudo bash $bash_options ./"$PACKAGE_SCRIPT"
		else
			bash $bash_options ./"$PACKAGE_SCRIPT"
		fi
	fi
fi
# END:   package install script

# BEGIN: set back system time
clear
while true; do
	if [ -f "/run/${TIME_RUN_FILE}" ]; then
		# call this block once only
		echo "Die Systemzeit wurde offenbar schon zurückgesetzt; wird übersprungen."
		echo
		break
	fi
	cat <<EOT
AKTION ERFORDERLICH: Systemzeit zurücksetzen?
======================================================================

Zur Sicherheit sollte nun die Systemzeit zurückgesetzt werden
(in Software, nicht die Rechneruhr).

EOT
	if yesno "Systemzeit zurücksetzen?"; then
		if [ "0" -ne "$EUID" ]; then
			echo "Jetzt müssen Sie möglicherweise das root-Passwort eingeben:"
			sudo bash $bash_options ./"$TIME_SCRIPT"
		else
			bash $bash_options ./"$TIME_SCRIPT"
		fi
	fi
	echo
	break
done
# END:   set back system time

while true; do
	# BEGIN: select security level
	clear
	while true; do
		cat <<EOT
AKTION ERFORDERLICH: Schlüsseltyp auswählen
======================================================================

Möchten Sie 

	(1) einen Alltagsschlüssel (mit Offline-Hauptschlüssel)

oder

	(2) einen Hochsicherheits-Schlüssel

erzeugen? (Sie können natürlich mehrere Schlüssel nacheinander erzeugen.)

EOT
		echo -n "1 / 2: "
		read answer
		if [ 1 = "$answer" -o 2 = "$answer" ]; then
			if [ 1 = "$answer" ]; then
				mode="normal"
			else
				mode="hisec"
			fi
			echo
			break
		else
			echo "falsche Eingabe, noch mal..."
		fi
	done
	# END:   select security level

	# BEGIN: with ot without policy URL?
	ask_policy_url
	# END:   with ot without policy URL?

	# BEGIN: run UID script
	clear
	while true; do
		if [ "yes" = "$singlestep" ] && ! yesno "UID-Script ausführen?"; then
			echo
			echo
			break
		fi
		if [ normal = "$mode" ]; then
			parameter=(--no-intro "$policy_url_parameter")
		else
			parameter=(--no-intro --high-security)
		fi
		if [ -x "./${USERID_SCRIPT}" ]; then
			if ! bash $bash_options "./${USERID_SCRIPT}" ${parameter[@]}; then
				echo "Fehler bei der Ausführung des User-ID-Scripts; Abbruch."
				echo "Weiter mit Return..."
				read cont
				exit 1
			fi
		else
			echo "FEHLER: Das User-ID-Script existiert nicht; Abbruch."
			echo "Weiter mit Return..."
			read cont
			exit 1
		fi
		echo
		if yesno "Sind Sie mit dieser Eingabe zufrieden?"; then
			break
		else
			continue
		fi
	done      
	# END:   run UID script
						
	# BEGIN: run policy URL script
	if [ -x "./${POLICYURL_SCRIPT}" ]; then
		while true; do
			if [ "yes" = "$singlestep" ] && ! yesno "Policy-URL--Script ausführen?"; then
				break
			fi
			if ! bash $bash_options "./${POLICYURL_SCRIPT}" "$policy_url_parameter"; then
				echo "Fehler bei der Ausführung des Policy-URL-Scripts."
				if ! yesno "Noch mal ausführen?"; then
					break
				fi
			else
				break
			fi
		done
	else
		echo "FEHLER: Das Policy-URL-Script existiert nicht; Abbruch."
		echo "Weiter mit Return..."
		read cont
		exit 1
	fi
	# END:   run policy URL script

	index=0
	if [ normal = "$mode" ]; then
		parameter=()
	else
		parameter=(--high-security)
		((index++))
	fi

	start_keygen_script () {
		bash $bash_options "./${KEYGEN_SCRIPT}" "$@"
	} # start_keygen_script ()

	if [ -x "./${KEYGEN_SCRIPT}" ]; then
		if ! start_keygen_script "${parameter[@]}"; then
			echo "Fehler bei der Ausführung des Schlüsselerzeugungs-Scripts; Abbruch."
			echo "Weiter mit Return..."
			read cont
		else
			if yesno "Noch einen Schlüssel erzeugen?"; then
				echo
				echo "Speichern Sie bitte erst die Schlüsseldateien."
				echo "Sorgen Sie dann dafür, dass das Verzeichnis 'export' leer ist."
				echo
				echo "Weiter mit Return..."
				read cont
				continue
			else
				break
			fi
		fi
	else
		echo "FEHLER: Das Schlüsselerzeugungs-Script existiert nicht; Abbruch."
		exit 1
	fi
done


